from ytree.frontends.rockstar import RockstarArbor
from ytree.testing.arbor_test import ArborTest
from ytree.testing.utilities import TempDirTest


class RockstarArborTest(TempDirTest, ArborTest):
    arbor_type = RockstarArbor
    test_filename = "rockstar/rockstar_halos/out_0.list"
    num_data_files = 65
    tree_skip = 10



import numpy as np
import pytest
import sys

from numpy.testing import assert_array_equal
from unittest import skipIf

from ytree.data_structures.load import load

try:
    from mpi4py import MPI
except ModuleNotFoundError:
    MPI = None


class ParallelTest:
    """
    Test class for ytree parallelism.
    """

    base_filename = "tiny_ctrees/locations.dat"
    test_filename = "test_arbor/test_arbor.h5"
    test_script = None
    ncores = 4

    def check_values(self, arbor):
        group = self.args[1]

        if "nodes" in group:
            increment = int(group.split("-")[1])
        else:
            increment = None

        assert_array_equal(arbor["test_field"], 2 * arbor["mass"])
        for tree in arbor:
            if increment is None:
                assert_array_equal(tree[group, "test_field"], 2 * tree[group, "mass"])

            else:
                all_inds = np.arange(tree.tree_size)
                my_inds = all_inds[::increment]
                select = np.isin(all_inds, my_inds)
                assert_array_equal(
                    tree["forest", "test_field"][select],
                    2 * tree["forest", "mass"][select],
                )

                # the ones we didn't do should all be -1
                assert_array_equal(
                    tree["forest", "test_field"][~select],
                    -np.ones_like(tree["forest", "test_field"][~select]),
                )

    @skipIf(MPI is None, "mpi4py not installed")
    @pytest.mark.parallel
    def test_parallel(self):
        args = [self.test_script, self.base_filename, self.test_filename] + [
            str(arg) for arg in self.args
        ]
        comm = MPI.COMM_SELF.Spawn(sys.executable, args=args, maxprocs=self.ncores)
        comm.Barrier()
        comm.Disconnect()

        test_arbor = load(self.test_filename)
        self.check_values(test_arbor)



"""
amiga halo finder frontend



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from ytree.frontends.ahf.arbor import AHFArbor, AHFCRMArbor



"""
tests for plotting



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from ytree.testing.utilities import requires_file, TempDirTest

import ytree

CT = "consistent_trees/tree_0_0_0.dat"


class TreePlotTest(TempDirTest):
    @requires_file(CT)
    def test_default_plot(self):
        a = ytree.load(CT)
        p = ytree.TreePlot(a[0])
        p.save()

    @requires_file(CT)
    def test_non_defaults(self):
        attrs = {
            "size_field": "virial_radius",
            "size_log": False,
            "min_mass": 1e14,
            "min_mass_ratio": 0.1,
        }

        a = ytree.load(CT)

        for attr, val in attrs.items():
            p = ytree.TreePlot(a[0])
            setattr(p, attr, val)
            p.save()

    @requires_file(CT)
    def test_save(self):
        a = ytree.load(CT)
        p = ytree.TreePlot(a[0])
        p.save("tree.png")

    @requires_file(CT)
    def test_dot_kwargs(self):
        a = ytree.load(CT)
        p = ytree.TreePlot(a[0], dot_kwargs={"dpi": 200})
        p.save()

    @requires_file(CT)
    def test_node_function(self):
        def my_func(halo):
            label = f"{halo['uid']}"
            return {"label": label}

        a = ytree.load(CT)
        p = ytree.TreePlot(a[0], node_function=my_func)
        p.save()

    @requires_file(CT)
    def test_node_function_bad(self):
        a = ytree.load(CT)
        with self.assertRaises(RuntimeError):
            ytree.TreePlot(a[0], node_function="notafunc")

    @requires_file(CT)
    def test_edge_function(self):
        def my_func(desc, anc):
            return {"color": "red"}

        a = ytree.load(CT)
        p = ytree.TreePlot(a[0], edge_function=my_func)
        p.save()

    @requires_file(CT)
    def test_edge_function_bad(self):
        a = ytree.load(CT)
        with self.assertRaises(RuntimeError):
            ytree.TreePlot(a[0], edge_function="notafunc")



import pytest


def is_parallel(item):
    if "parallel" in item.keywords:
        return True
    else:
        if getattr(item, "cls", None) is None:
            return False
        if getattr(item.cls, "ncores", 1) > 1:
            return True
    return False


def is_slow(item):
    if getattr(item, "cls", None) is None:
        return False
    if getattr(item.cls, "slow", False):
        return True
    return False


def pytest_addoption(parser):
    parser.addoption(
        "--exclude-slow",
        action="store_true",
        default=False,
        help="exclude tests marked as slow",
    )
    parser.addoption(
        "--exclude-serial",
        action="store_true",
        default=False,
        help="exclude tests not using parallelism",
    )
    parser.addoption(
        "--exclude-parallel",
        action="store_true",
        default=False,
        help="exclude tests using parallelism",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "parallel: test uses parallelism")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--exclude-slow"):
        skip_slow = pytest.mark.skip(reason="excluding tests marked as slow")
        for item in items:
            if is_slow(item):
                item.add_marker(skip_slow)

    if config.getoption("--exclude-serial"):
        skip_serial = pytest.mark.skip(reason="excluding tests not using parallelism")
        for item in items:
            if not is_parallel(item):
                item.add_marker(skip_serial)

    if config.getoption("--exclude-parallel"):
        skip_parallel = pytest.mark.skip(reason="excluding tests using parallelism")
        for item in items:
            if is_parallel(item):
                item.add_marker(skip_parallel)



"""
testing utilities



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import h5py
import numpy as np
import os
import shutil
import subprocess
import tempfile
from unittest import TestCase

from numpy.dtypes import StringDType
from numpy.testing import assert_equal, assert_almost_equal, assert_array_equal
from yt.funcs import get_pbar

from ytree.utilities.loading import get_path
from ytree.utilities.logger import ytreeLogger as mylog


def requires_file(filename):
    def ffalse(func):
        return None

    def ftrue(func):
        return func

    if not isinstance(filename, list):
        filename = [filename]
    try:
        [get_path(fn) for fn in filename]
    except IOError:
        return ffalse
    return ftrue


class TempDirTest(TestCase):
    """
    A test class that runs in a temporary directory and
    removes it afterward.
    """

    def setUp(self):
        self.curdir = os.getcwd()
        self.tmpdir = tempfile.mkdtemp()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.curdir)
        shutil.rmtree(self.tmpdir)


def run_command(command, timeout=None):
    try:
        proc = subprocess.run(command, shell=True, timeout=timeout)
        if proc.returncode == 0:
            success = True
        else:
            success = False
    except subprocess.TimeoutExpired:
        print("Process reached timeout of %d s. (%s)" % (timeout, command))
        success = False
    except KeyboardInterrupt:
        print("Killed by keyboard interrupt!")
        success = False
    return success


def get_tree_split(arbor):
    """
    Get a few separate ancestors from a tree.
    """
    ancs = None
    for tree in arbor:
        for halo in tree["tree"]:
            ancs = list(halo.ancestors)
            if len(ancs) > 1:
                break

    if ancs is None:
        raise RuntimeError("Could not find nodes to test with.")
    return ancs


def get_random_trees(arbor, seed, n):
    """
    Get n random trees from the arbor.
    """

    gen = np.random.default_rng(708)
    itrees = np.arange(arbor.size)
    gen.shuffle(itrees)
    for itree in itrees[:5]:
        yield arbor[itree]


def get_stringsafe_compare_arrays(arr1, arr2):
    """
    Convert to stringy arrays to a common type for comparison.
    """

    if StringDType() in (arr1.dtype, arr2.dtype):
        # Python 3.9 and earlier require this. Otherwise, you end up
        # with  "b'Alexander'" instead of 'Alexander'.
        if arr1.dtype != StringDType():
            arr1 = arr1.astype(str)
        arr1 = arr1.astype(StringDType())
        if arr2.dtype != StringDType():
            arr2 = arr2.astype(str)
        arr2 = arr2.astype(StringDType())

    return arr1, arr2


def compare_arbors(a1, a2, groups=None, fields=None, skip1=1, skip2=1):
    """
    Compare all fields for all trees in two arbors.
    """

    if groups is None:
        groups = ["tree", "prog"]

    if fields is None:
        fields = a1.field_list

    for i, field in enumerate(fields):
        c1, c2 = get_stringsafe_compare_arrays(a1[field][::skip1], a2[field][::skip2])
        mylog.info(f"Comparing arbor field: {field} ({i + 1}/{len(fields)}).")
        assert_array_equal(c1, c2, err_msg=f"Arbor field mismatch: {a1, a2, field}.")

    trees1 = list(a1[::skip1])
    trees2 = list(a2[::skip2])

    ntot = len(trees1)
    pbar = get_pbar("Comparing trees", ntot)
    for i, (t1, t2) in enumerate(zip(trees1, trees2)):
        compare_trees(t1, t2, groups=groups, fields=fields)
        pbar.update(i + 1)
    pbar.finish()


def compare_trees(t1, t2, groups=None, fields=None):
    """
    Compare all fields between two trees.
    """

    if groups is None:
        groups = ["tree", "prog"]

    if fields is None:
        fields = t1.arbor.field_list

    for field in fields:
        for group in groups:
            c1, c2 = get_stringsafe_compare_arrays(t1[group, field], t2[group, field])
            assert_array_equal(
                c1, c2, err_msg=f"Tree comparison failed for {group} field: {field}."
            )
    t1.arbor.reset_node(t1)
    t2.arbor.reset_node(t2)


def compare_hdf5(fh1, fh2, compare=None, compare_groups=True, **kwargs):
    """
    Compare all datasets between two hdf5 files.
    """

    if compare is None:
        compare = assert_array_equal
    if not isinstance(fh1, h5py.Group):
        fh1 = h5py.File(fh1, "r")
    if not isinstance(fh2, h5py.Group):
        fh2 = h5py.File(fh2, "r")

    if compare_groups:
        err_msg = f"{fh1.file.filename} and {fh2.file.filename} have different datasets in group {fh1.name}."
        assert_equal(
            sorted(list(fh1.keys())), sorted(list(fh2.keys())), err_msg=err_msg
        )

    for key in fh1.keys():
        if isinstance(fh1[key], h5py.Group):
            compare_hdf5(
                fh1[key],
                fh2[key],
                compare_groups=compare_groups,
                compare=compare,
                **kwargs,
            )
        else:
            err_msg = (
                f"{key} field not equal for {fh1.file.filename} and {fh2.file.filename}"
            )
            if fh1[key].dtype == "int":
                assert_array_equal(fh1[key][()], fh2[key][()], err_msg=err_msg)
            else:
                compare(fh1[key][()], fh2[key][()], err_msg=err_msg, **kwargs)


def assert_rel_equal(a1, a2, decimals, err_msg="", verbose=True):
    # We have nan checks in here because occasionally we have fields that get
    # weighted without non-zero weights.  I'm looking at you, particle fields!
    if isinstance(a1, np.ndarray):
        assert a1.size == a2.size
        # Mask out NaNs
        assert (np.isnan(a1) == np.isnan(a2)).all()
        a1[np.isnan(a1)] = 1.0
        a2[np.isnan(a2)] = 1.0
        # Mask out 0
        ind1 = np.array(np.abs(a1) < np.finfo(a1.dtype).eps)
        ind2 = np.array(np.abs(a2) < np.finfo(a2.dtype).eps)
        assert (ind1 == ind2).all()
        a1[ind1] = 1.0
        a2[ind2] = 1.0
    elif np.any(np.isnan(a1)) and np.any(np.isnan(a2)):
        return True
    if not isinstance(a1, np.ndarray) and a1 == a2 == 0.0:
        # NANS!
        a1 = a2 = 1.0
    return assert_almost_equal(
        np.array(a1) / np.array(a2), 1.0, decimals, err_msg=err_msg, verbose=verbose
    )


def assert_array_rel_equal(a1, a2, decimals=16, **kwargs):
    """
    Wraps assert_rel_equal with, but decimals is a keyword arg.
    """
    assert_rel_equal(a1, a2, decimals, **kwargs)


def verify_get_node(my_tree, n=3):
    """
    Unit tests for get_node.
    """
    for selector in ["forest", "tree", "prog"]:
        node_list = list(my_tree[selector])

        inodes = np.arange(len(node_list))
        gen = np.random.default_rng(847)
        gen.shuffle(inodes)

        for inode in inodes[:3]:
            my_node = my_tree.get_node(selector, inode)
            err_msg = f"get_node failed: {selector} " + f"with {str(my_tree.arbor)}."
            assert_equal(my_node.uid, node_list[inode].uid, err_msg=err_msg)

            if selector == "forest":
                err_msg = (
                    "get_node index is not tree_id for " + f"{str(my_tree.arbor)}."
                )
                assert_equal(my_node.tree_id, inode, err_msg=err_msg)


def verify_get_leaf_nodes(my_tree):
    """
    Unit tests for get_leaf_nodes.
    """
    for selector in ["forest", "tree", "prog"]:
        uids1 = np.array(
            [node.uid for node in my_tree.get_leaf_nodes(selector=selector)]
        )
        uids2 = np.array(
            [
                my_halo.uid
                for my_halo in my_tree[selector]
                if not list(my_halo.ancestors)
            ]
        )

        err_msg = f"get_leaf_nodes failure for {selector} in {my_tree.arbor}."
        assert_equal(uids1, uids2, err_msg=err_msg)


def verify_get_root_nodes(my_tree):
    """
    Unit tests for get_root_nodes.
    """

    root_nodes1 = list(my_tree.get_root_nodes())
    for root_node in root_nodes1:
        assert_equal(
            root_node["desc_uid"],
            -1,
            err_msg=f"In {my_tree.arbor}: {root_node} has "
            + f"desc_uid={root_node['desc_uid']}, but expected -1.",
        )

    root_nodes2 = [node for node in my_tree["forest"] if node.descendent is None]

    uids1 = np.sort([node.uid for node in root_nodes1])
    uids2 = np.sort([node.uid for node in root_nodes2])

    assert_array_equal(
        uids1, uids2, err_msg=f"get_root_nodes failure in {my_tree.arbor}."
    )



"""
tests for the plotting examples in the docs



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from ytree.testing.utilities import requires_file, TempDirTest

import ytree


def my_node(halo):
    prog = list(halo.find_root()["prog", "uid"])
    if halo["uid"] in prog:
        color = "red"
    else:
        color = "black"

    label = f"""
    id: {halo["uid"]:d}
    mass: {halo["mass"].to("Msun"):.2e} Msun
    """

    my_kwargs = {"label": label, "fontsize": 8, "shape": "square", "color": color}
    return my_kwargs


def my_edge(ancestor, descendent):
    if descendent["mass"] < ancestor["mass"]:
        color = "blue"
    elif descendent["mass"] / ancestor["mass"] > 10:
        color = "green"
    else:
        color = "black"

    my_kwargs = {"color": color, "penwidth": 5}
    return my_kwargs


AHF = "ahf_halos/snap_N64L16_000.parameter"


class TreePlotTest(TempDirTest):
    @requires_file(AHF)
    def test_tree_plot(self):
        a = ytree.load(AHF, hubble_constant=0.7)
        p = ytree.TreePlot(a[0], dot_kwargs={"rankdir": "LR", "size": '"12,4"'})
        p.save("tree.png")

    @requires_file(AHF)
    def test_tree_plot_small(self):
        a = ytree.load(AHF, hubble_constant=0.7)
        p = ytree.TreePlot(a[0], dot_kwargs={"rankdir": "LR", "size": '"12,4"'})
        p.min_mass_ratio = 0.01
        p.save("tree_small.png")

    @requires_file(AHF)
    def test_tree_plot_custom_node(self):
        a = ytree.load(AHF, hubble_constant=0.7)
        p = ytree.TreePlot(a[0], dot_kwargs={"rankdir": "BT"}, node_function=my_node)
        p.save("tree_custom_node.png")

    @requires_file(AHF)
    def test_tree_plot_custom_edge(self):
        a = ytree.load(AHF, hubble_constant=0.7)
        p = ytree.TreePlot(
            a[0],
            dot_kwargs={"rankdir": "BT"},
            node_function=my_node,
            edge_function=my_edge,
        )
        p.save("tree_custom_edge.png")



from ytree.frontends.consistent_trees_hdf5 import ConsistentTreesHDF5Arbor
from ytree.testing.arbor_test import ArborTest
from ytree.testing.utilities import TempDirTest


class ConsistentTreesHDF5ArborTest1(TempDirTest, ArborTest):
    arbor_type = ConsistentTreesHDF5Arbor
    test_filename = "consistent_trees_hdf5/soa/forest.h5"
    num_data_files = 1
    tree_skip = 10000



from ytree.frontends.gadget4 import Gadget4Arbor
from ytree.testing.arbor_test import ArborTest
from ytree.testing.utilities import TempDirTest


class Gadget4ArborMultipleTest(TempDirTest, ArborTest):
    arbor_type = Gadget4Arbor
    test_filename = "gadget4/treedata/trees.0.hdf5"
    groups = ("forest", "tree", "prog")
    num_data_files = 64
    tree_skip = 100



"""
ytree imports



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from ytree.analysis.analysis_pipeline import AnalysisPipeline
from ytree.data_structures.load import load
from ytree.data_structures.tree_node_selector import (
    TreeNodeSelector,
    add_tree_node_selector,
)
from ytree.visualization.tree_plot import TreePlot

from ytree.utilities.parallel import parallel_trees, parallel_tree_nodes, parallel_nodes

__version__ = "3.3.1"



"""
parallel_nodes test script



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import ytree
from ytree.testing.utilities import get_tree_split
from mpi4py import MPI
import sys
import yt

yt.enable_parallelism()


def run():
    input_fn, output_fn, selection, group = sys.argv[1:5]
    njobs = tuple([int(arg) for arg in sys.argv[5:7]])
    dynamic = tuple([bool(int(arg)) for arg in sys.argv[7:9]])
    if len(sys.argv) > 9:
        save_every = int(sys.argv[9])
    else:
        save_every = None

    a = ytree.load(input_fn)
    if "test_field" not in a.field_list:
        a.add_analysis_field("test_field", default=-1, units="Msun")

    if selection == "all":
        trees = list(a[:8])
    elif selection == "nonroot":
        trees = get_tree_split(a)
    else:
        print(f"Bad selection: {selection}.")
        sys.exit(1)

    for node in ytree.parallel_nodes(
        trees,
        group=group,
        njobs=njobs,
        dynamic=dynamic,
        save_every=save_every,
        filename=output_fn,
    ):
        root = node.root
        yt.mylog.info(f"Doing {node.tree_id}/{root.tree_size} of {root._arbor_index}")
        node["test_field"] = 2 * node["mass"]


if __name__ == "__main__":
    comm = MPI.Comm.Get_parent()
    try:
        run()
    except BaseException as e:
        print(f"Exception raised on {comm.rank} of {comm.size}: {e}")
    comm.Barrier()
    comm.Disconnect()



import os
import sys

from ytree.testing.utilities import run_command
from ytree.utilities.io import dirname
from ytree.utilities.loading import check_path

try:
    from mpi4py import MPI
except ModuleNotFoundError:
    MPI = None


class ExampleScriptTest:
    """
    Tests for the code examples.
    """

    script_filename = None
    input_filename = None
    timeout = 60
    output_files = ()
    ncores = 1

    def test_example(self):
        if self.script_filename is None:
            return

        if self.input_filename is not None:
            try:
                check_path(self.input_filename)
            except IOError:
                self.skipTest("test file missing")

        source_dir = dirname(__file__, level=3)
        script_path = os.path.join(
            source_dir, "doc", "source", "examples", self.script_filename
        )

        if self.ncores > 1:
            if MPI is None:
                self.skipTest("mpi4py not installed")
            comm = MPI.COMM_SELF.Spawn(
                sys.executable, args=[script_path], maxprocs=self.ncores
            )
            comm.Barrier()
            comm.Disconnect()
        else:
            command = f"{sys.executable} {script_path}"
            assert run_command(command, timeout=self.timeout)

        for fn in self.output_files:
            assert os.path.exists(fn)



from ytree.frontends.consistent_trees import ConsistentTreesArbor
from ytree.testing.arbor_test import ArborTest
from ytree.testing.utilities import TempDirTest


class ConsistentTreesArborTest(TempDirTest, ArborTest):
    arbor_type = ConsistentTreesArbor
    test_filename = "consistent_trees/tree_0_0_0.dat"
    num_data_files = 1



"""
TreeFrogArbor class and member functions



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import h5py
import numpy as np
import os
import re

from ytree.data_structures.arbor import SegmentedArbor
from ytree.frontends.treefrog.fields import TreeFrogFieldInfo
from ytree.frontends.treefrog.io import (
    TreeFrogDataFile,
    TreeFrogRootFieldIO,
    TreeFrogTreeFieldIO,
)
from ytree.utilities.logger import ytreeLogger as mylog


class TreeFrogArbor(SegmentedArbor):
    """
    Arbors loaded from consistent-trees data converted into HDF5.
    """

    _suffix = ".hdf5"
    _default_dtype = None
    _field_info_class = TreeFrogFieldInfo
    _root_field_io_class = TreeFrogRootFieldIO
    _tree_field_io_class = TreeFrogTreeFieldIO

    def __init__(self, filename):
        filename = self._determine_filename(filename)
        super().__init__(filename)

    def _determine_filename(self, filename):
        """
        Find the proper parameter filename if given a data file instead.
        """

        suffix = f".foreststats{self._suffix}"
        if filename.endswith(suffix):
            return filename

        match = re.search(rf"{self._suffix}\.\d+$", filename)
        forest_fn = filename[: match.start()] + suffix
        return forest_fn

    def _get_data_files(self):
        self.data_files = [
            TreeFrogDataFile(f"{self._prefix}{self._suffix}.{i}")
            for i in range(self._nfiles)
        ]

    @property
    def _prefix(self):
        suffix = f".foreststats{self._suffix}"
        return self.parameter_filename[: -len(suffix)]

    def _parse_parameter_file(self):
        with h5py.File(self.parameter_filename, mode="r") as f:
            self._nfiles = f["Header"].attrs["NFiles"]
            self._nsnaps = f["Header"].attrs["NSnaps"]
            self._size = f["ForestInfo"].attrs["NForests"]
            self._file_count = f["ForestInfo"]["NForestsPerFile"][()]

        if self._nfiles < 1:
            mylog.warning(f"Dataset {self.parameter_filename} has no data files.")
            return

        fn = f"{self._prefix}{self._suffix}.0"
        if not os.path.exists(fn):
            raise RuntimeError(f"Data file not found: {fn}.")

        with h5py.File(fn, mode="r") as f:
            self.hubble_constant = f["Header/Simulation"].attrs["h_val"]
            self.omega_matter = f["Header/Simulation"].attrs["Omega_m"]
            self.omega_lambda = f["Header/Simulation"].attrs["Omega_Lambda"]
            self.box_size = self.quan(f["Header/Simulation"].attrs["Period"], "Mpc/h")
            if self._nsnaps < 1:
                mylog.warning(f"Dataset {self.parameter_filename} has no snapshots.")
                return

            self.units = {}
            for attr in [
                "Length_unit_to_kpc",
                "Mass_unit_to_solarmass",
                "Velocity_unit_to_kms",
            ]:
                self.units[attr] = f["Header/Units"].attrs[attr]

            field_list = list(f["Snap_000"].keys())
            self._scale_factors = np.array(
                [f[f"Snap_{i:03d}"].attrs["scalefactor"] for i in range(self._nsnaps)]
            )

        self.field_list = field_list
        self.field_info.update({field: {} for field in field_list})

        self.field_list.append("scale_factor")
        self.field_info["scale_factor"] = {"source": "arbor"}

    def _plant_trees(self):
        if self.is_planted or self._size == 0:
            return

        with h5py.File(self.parameter_filename, mode="r") as f:
            self._node_info["_tree_size"] = f["ForestInfo"]["ForestSizes"][()]

        ei = self._file_count.cumsum()
        si = ei - self._file_count
        trees = np.arange(self._size)
        fi = np.digitize(trees, ei)
        # number of the file each forest lives in
        self._node_info["_fi"] = fi
        # index within that file each forest is at
        self._node_info["_si"] = trees - si[fi]
        # This is slightly unsafe since the call to get_fields to get the uid field
        # will call _plant_trees, too. It will not get here because is_planted will
        # be True. As long as _fi and _si are initialized properly, it should be ok.
        self._node_info["uid"] = self["uid"]

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        Should be an hdf5 file with a few key attributes.
        """
        fn = self._determine_filename(self, args[0])
        suffix = f".foreststats{self._suffix}"
        if not fn.endswith(suffix):
            return False
        if not h5py.is_hdf5(fn):
            return False

        info = {
            "Header": [
                "FileSplittingCriteria",
                "FilesSplitByForest",
                "HaloCatalogBaseFileName",
                "NFiles",
                "NSnaps",
            ],
            "ForestInfo": [
                "MaxForestFOFGroupID",
                "MaxForestFOFGroupSize",
                "MaxForestID",
                "MaxForestSize",
                "NForests",
            ],
        }

        with h5py.File(fn, mode="r") as f:
            for group, attrs in info.items():
                if group not in f:
                    return False
                for attr in attrs:
                    if attr not in f[group].attrs:
                        return False
        return True



from ytree.frontends.ahf import AHFArbor
from ytree.testing.arbor_test import ArborTest
from ytree.testing.utilities import TempDirTest


class AHFNameConfigArborTest(TempDirTest, ArborTest):
    arbor_type = AHFArbor
    test_filename = "B25_N256_CDM_1LPT/AHF.B25_N256_CDM_1LPT.snap_055.parameter"
    load_kwargs = {
        "name_config": {
            "ahf_prefix": "AHF.B25_N256_CDM_1LPT",
            "mtree_prefix": "MTREE.B25_N256_CDM_1LPT.z39_adapt",
            "mtree_suffix": ".AHF_croco",
        }
    }
    num_data_files = 2
    tree_skip = 100



"""
tests for parallel_trees iterator



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import os

from ytree.testing.parallel_test import ParallelTest
from ytree.testing.utilities import TempDirTest

script_path = os.path.dirname(__file__)
test_script = os.path.join(script_path, "run_parallel_trees.py")


class ParallelTreesTest2(TempDirTest, ParallelTest):
    test_script = test_script
    args = ("all", "tree", 0, 0, 4)



import numpy as np
from numpy.testing import assert_equal, assert_array_equal

from ytree.data_structures.load import load
from ytree.frontends.ytree import YTreeArbor
from ytree.testing.utilities import (
    compare_arbors,
    get_random_trees,
    verify_get_leaf_nodes,
    verify_get_node,
    verify_get_root_nodes,
)
from ytree.utilities.loading import get_path
from ytree.utilities.logger import ytreeLogger as mylog


class ArborTest:
    """
    A battery of tests for all frontends.
    """

    arbor_type = None
    test_filename = None
    load_kwargs = None
    load_callback = None
    groups = ("tree", "prog")
    num_data_files = None
    tree_skip = 1
    custom_vector_fields = None
    slow = False

    _arbor = None

    @property
    def arbor(self):
        if self._arbor is None:
            try:
                self.test_filename = get_path(self.test_filename)
            except IOError:
                self.skipTest("test file missing")

            if self.load_kwargs is None:
                self.load_kwargs = {}

            self._arbor = load(self.test_filename, **self.load_kwargs)
            if self.load_callback is not None:
                self.load_callback(self._arbor)
        return self._arbor

    def test_arbor_type(self):
        assert isinstance(self.arbor, self.arbor_type)

    def test_data_files(self):
        if self.num_data_files is None:
            return
        assert_equal(
            len(self.arbor.data_files),
            self.num_data_files,
            err_msg=f"Incorrect number of data files for {self.arbor}.",
        )

    def test_get_root_nodes(self):
        for my_tree in get_random_trees(self.arbor, 47457, 5):
            verify_get_root_nodes(my_tree)

    def test_get_roor_nodes_nonroot(self):
        my_tree = list(self.arbor[0].ancestors)[0]
        verify_get_root_nodes(my_tree)

    def test_get_leaf_nodes(self):
        for my_tree in get_random_trees(self.arbor, 41153, 5):
            verify_get_leaf_nodes(my_tree)

    def test_get_leaf_nodes_ungrown_nonroot(self):
        my_tree = list(self.arbor[0].ancestors)[0]
        verify_get_leaf_nodes(my_tree)

    def test_get_node(self):
        for my_tree in get_random_trees(self.arbor, 47988, 5):
            verify_get_node(my_tree)

            ihalos = np.arange(1, my_tree.tree_size)
            gen = np.random.default_rng(312)
            gen.shuffle(ihalos)
            for ihalo in ihalos[:3]:
                my_halo = my_tree.get_node("forest", ihalo)
                verify_get_node(my_halo)

    def test_get_node_ungrown_nonroot(self):
        my_tree = list(self.arbor[0].ancestors)[0]
        my_halo = my_tree.get_node("forest", 0)
        node_list = list(my_tree["forest"])
        assert_equal(my_halo.uid, node_list[0].uid)

    def test_reset_node(self):
        t = self.arbor[0]
        ts0 = len(list(t["tree"]))
        f0 = dict((field, t["tree", field]) for field in ["uid", "desc_uid"])

        assert self.arbor.is_setup(t)
        assert self.arbor.is_grown(t)

        self.arbor.reset_node(t)

        for attr in self.arbor._reset_attrs:
            assert getattr(t, attr) is None
        assert_equal(len(t.field_data), 0)
        assert not self.arbor.is_setup(t)
        assert not self.arbor.is_grown(t)

        assert_equal(
            len(list(t["tree"])),
            ts0,
            err_msg=f"Trees are not the same size after resetting for {self.arbor}.",
        )

        for field in f0:
            assert_array_equal(
                t["tree", field],
                f0[field],
                err_msg=f"Tree field {field} not the same after resetting for {self.arbor}.",
            )

    def test_reset_nonroot(self):
        t = self.arbor[0]
        node = list(t["tree"])[1]
        ts0 = len(list(node["tree"]))
        f0 = dict((field, node["tree", field]) for field in ["uid", "desc_uid"])

        self.arbor.reset_node(node)

        assert_equal(
            len(list(node["tree"])),
            ts0,
            err_msg=f"Trees are not the same size after resetting for {self.arbor}.",
        )

        for field in f0:
            assert_array_equal(
                node["tree", field],
                f0[field],
                err_msg=f"Tree field {field} not the same after resetting for {self.arbor}.",
            )

    def test_save_and_reload(self):
        skip = self.tree_skip
        if skip > 1:
            trees = list(self.arbor[::skip])
        else:
            trees = None

        fn = self.arbor.save_arbor(trees=trees)

        save_arbor = load(fn)
        if self.load_callback is not None:
            self.load_callback(save_arbor)

        assert isinstance(save_arbor, YTreeArbor)
        compare_arbors(save_arbor, self.arbor, groups=self.groups, skip2=skip)

    def test_vector_fields(self):
        a = self.arbor
        t = a[0]

        if self.custom_vector_fields is not None:
            for vfield, cfields in self.custom_vector_fields:
                a.add_vector_field(vfield, vector_components=cfields)

        for field in a.field_info.vector_fields:
            mylog.info(f"Comparing vector field: {field}.")
            magfield = np.sqrt((a[field] ** 2).sum(axis=1))
            assert_array_equal(
                a[f"{field}_magnitude"],
                magfield,
                err_msg=f"Magnitude field incorrect: {field}.",
            )

            cfields = a.field_info[field]["vector_components"]

            for i, cfield in enumerate(cfields):
                assert_array_equal(
                    a[cfield],
                    a[field][:, i],
                    err_msg=(
                        f"Arbor vector field {field} does not match in dimension {i}."
                    ),
                )

                assert_array_equal(
                    t[cfield],
                    t[field][i],
                    err_msg=(
                        f"Tree vector field {field} does not match in dimension {i}."
                    ),
                )

                for group in ["prog", "tree"]:
                    assert_array_equal(
                        t[group, cfield],
                        t[group, field][:, i],
                        err_msg=(
                            f"{group} vector field {field} does not match "
                            f"in dimension {i}."
                        ),
                    )



"""
tests for example scripts



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from ytree.testing.example_script_test import ExampleScriptTest
from ytree.testing.utilities import TempDirTest


class TestPlotMostMassive(TempDirTest, ExampleScriptTest):
    script_filename = "plot_most_massive.py"
    input_filename = "tiny_ctrees/locations.dat"
    output_files = ("most_massive.png",)


class TestPlotMostHalos(TempDirTest, ExampleScriptTest):
    script_filename = "plot_most_halos.py"
    input_filename = "tiny_ctrees/locations.dat"
    output_files = ("most_halos.png",)



from ytree.frontends.csv import CSVArbor
from ytree.testing.arbor_test import ArborTest
from ytree.testing.utilities import TempDirTest


class CSVArborTest(TempDirTest, ArborTest):
    arbor_type = CSVArbor
    test_filename = "csv/trees.csv"

    def load_callback(self, a):
        a.set_selector("max_field_value", "charisma")



from ytree.frontends.consistent_trees import ConsistentTreesHlistArbor
from ytree.testing.arbor_test import ArborTest
from ytree.testing.utilities import TempDirTest


class ConsistentTreesHlistArborTest(TempDirTest, ArborTest):
    arbor_type = ConsistentTreesHlistArbor
    test_filename = "ctrees_hlists/hlists/hlist_0.12521.list"
    num_data_files = 10
    tree_skip = 100



from ytree.frontends.consistent_trees_hdf5 import ConsistentTreesHDF5Arbor
from ytree.testing.arbor_test import ArborTest
from ytree.testing.utilities import TempDirTest


class ConsistentTreesHDF5ArborTest4(TempDirTest, ArborTest):
    arbor_type = ConsistentTreesHDF5Arbor
    test_filename = [
        "consistent_trees_hdf5/soa/forest_0.h5",
        "consistent_trees_hdf5/soa/forest_0.h5",
    ]
    num_data_files = 2
    tree_skip = 20000



"""
tests for derived fields



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal, assert_equal

from ytree.testing.utilities import requires_file, TempDirTest

import ytree

CT = "consistent_trees/tree_0_0_0.dat"


def _potential(field, data):
    return data["mass"] / data["virial_radius"]


def _uid_mod7(field, data):
    return data["uid"] % 7


class DerivedFieldTest(TempDirTest):
    @requires_file(CT)
    def test_derived_fields(self):
        a = ytree.load(CT)

        a.add_derived_field("potential", _potential, units="Msun/kpc")

        m = a["mass"].copy()
        r = a["virial_radius"].copy()
        assert_array_equal(a["potential"], m / r)
        assert_equal(str(a["potential"].units), "Msun/kpc")
        assert_equal(a["potential"].dtype, a._default_dtype)

        my_tree = a[0]
        m = my_tree["mass"].copy()
        r = my_tree["virial_radius"].copy()
        assert_array_equal(my_tree["potential"], m / r)
        assert_equal(str(my_tree["potential"].units), "Msun/kpc")
        assert_equal(my_tree["potential"].dtype, a._default_dtype)

    @requires_file(CT)
    def test_derived_fields_int(self):
        a = ytree.load(CT)

        my_dtype = np.int64
        a.add_derived_field("uid_mod7", _uid_mod7, units="", dtype=my_dtype)

        um7 = a["uid"] % 7
        assert_array_equal(a["uid_mod7"], um7)
        assert_equal(a["uid_mod7"].dtype, my_dtype)

        my_tree = a[0]
        um7 = my_tree["tree", "uid"] % 7
        assert_array_equal(my_tree["tree", "uid_mod7"], um7)
        assert_equal(my_tree["tree", "uid_mod7"].dtype, my_dtype)



"""
ConsistentTreesHDF5Arbor fields



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from ytree.data_structures.fields import FieldInfoContainer

m_unit = "Msun"
p_unit = "unitary"
r_unit = "kpc"
v_unit = "km/s"


class ConsistentTreesHDF5FieldInfo(FieldInfoContainer):
    alias_fields = (
        ("uid", "id", ""),
        ("desc_uid", "desc_id", ""),
        ("halo_id", "Orig_halo_ID", ""),
        ("scale_factor", "scale", None),
        ("mass", "Mvir", m_unit),
        ("virial_mass", "Mvir", m_unit),
        ("virial_radius", "Rvir", r_unit),
        ("scale_radius", "rs", r_unit),
        ("velocity_dispersion", "vrms", v_unit),
        ("position_x", "x", p_unit),
        ("position_y", "y", p_unit),
        ("position_z", "z", p_unit),
        ("velocity_x", "vx", v_unit),
        ("velocity_y", "vy", v_unit),
        ("velocity_z", "vz", v_unit),
        ("angular_momentum_x", "Jx", None),
        ("angular_momentum_y", "Jy", None),
        ("angular_momentum_z", "Jz", None),
        ("spin_parameter", "Spin", None),
    )



"""
ytree-specific fields




"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from yt.fields.field_info_container import FieldInfoContainer

p_units = "unitary"
v_units = "km/s"


def _redshift(field, data):
    ptype, _ = field.name
    return 1.0 / data[ptype, "scale_factor"] - 1


def _time(field, data):
    ptype, _ = field.name
    return data.ds.cosmology.t_from_z(data[ptype, "redshift"])


class YTreeFieldInfo(FieldInfoContainer):
    known_other_fields = ()

    known_particle_fields = (
        ("position_x", (p_units, ("halos", "particle_position_x"), None)),
        ("position_y", (p_units, ("halos", "particle_position_y"), None)),
        ("position_z", (p_units, ("halos", "particle_position_z"), None)),
        ("velocity_x", (v_units, ("halos", "particle_velocity_x"), None)),
        ("velocity_y", (v_units, ("halos", "particle_velocity_y"), None)),
        ("velocity_z", (v_units, ("halos", "particle_velocity_z"), None)),
        ("mass", ("Msun", ("halos", "particle_mass"), None)),
    )

    def setup_particle_fields(self, ptype):
        if (ptype, "redshift") not in self.ds.field_list:
            self.add_field(
                (ptype, "redshift"),
                function=_redshift,
                sampling_type="particle",
                units="",
            )

        if (ptype, "time") not in self.ds.field_list:
            self.add_field(
                (ptype, "time"), function=_time, sampling_type="particle", units="Gyr"
            )

        super().setup_particle_fields(ptype)



"""
YTreeArbor utilities



"""

_ops = {"above": ">=", "below": "<=", "equal": "=="}


def get_conditional(op, criterion):
    my_op = _ops[op]

    if len(criterion) == 2:
        field, value = criterion
        units = None
    elif len(criterion) == 3:
        field, value, units = criterion
    else:
        raise ValueError(
            f"Criterion must have either two or three values: {criterion}."
        )

    if units is None:
        ustr = ""
    else:
        ustr = f'.in_units("{units}")'

    condition = f'obj["halos", "{field}"]{ustr} {my_op} {value}'
    return condition


def get_about(criterion):
    if len(criterion) == 3:
        field, value, within = criterion
        units = None
    elif len(criterion) == 4:
        field, value, units, within = criterion
    else:
        raise ValueError(
            f"Criterion must have either two or three values: {criterion}."
        )

    c1 = get_conditional("above", (field, (1 - within) * value, units))
    c2 = get_conditional("below", (field, (1 + within) * value, units))
    return [c1, c2]



"""
tests for example halo significance script



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from ytree.testing.example_script_test import ExampleScriptTest
from ytree.testing.utilities import TempDirTest


class TestHaloSignificance(TempDirTest, ExampleScriptTest):
    ncores = 2
    script_filename = "halo_significance.py"
    input_filename = "tiny_ctrees/locations.dat"
    output_files = (
        "halo_significance/halo_significance-analysis.h5",
        "halo_significance/halo_significance.h5",
        "halo_significance/halo_significance_0000-analysis.h5",
        "halo_significance/halo_significance_0000.h5",
    )



from ytree.frontends.ahf import AHFArbor
from ytree.testing.arbor_test import ArborTest
from ytree.testing.utilities import TempDirTest


class AHFArborTest(TempDirTest, ArborTest):
    arbor_type = AHFArbor
    test_filename = "ahf_halos/snap_N64L16_000.parameter"
    num_data_files = 136
    tree_skip = 100



"""
ytree exceptions



"""

# -----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------


class ArborDataFileEmpty(Exception):
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return f"Data file is empty: {self.filename}."


class ArborFieldException(Exception):
    def __init__(self, field, arbor=None):
        self.field = field
        self.arbor = arbor


class ArborDerivedFieldException(ArborFieldException):
    def __str__(self):
        return f'Derived field "{self.field}" could not be generated for {self.arbor}.'


class ArborFieldDependencyNotFound(Exception):
    def __init__(self, field, dependency, arbor=None):
        self.field = field
        self.dependency = dependency
        self.arbor = arbor

    def __str__(self):
        return (
            f'Field dependency not found: "{self.dependency}" '
            f'(dependency for "{self.field}") in {self.arbor}.'
        )


class ArborFieldCircularDependency(ArborFieldException):
    def __str__(self):
        return f'Field depends on itself: "{self.field}" in {self.arbor}.'


class ArborFieldNotFound(ArborFieldException):
    def __str__(self):
        return f'Field not found: "{self.field}" in {self.arbor}.'


class ArborFieldAlreadyExists(ArborFieldException):
    def __str__(self):
        return f'Field already exists: "{self.field}" in {self.arbor}.'


class ArborAnalysisFieldNotGenerated(ArborFieldException):
    def __str__(self):
        return (
            f'Analysis field "{self.field}" needed but '
            f"not yet generated in {self.arbor}."
        )


class ArborAnalysisFieldNotFound(ArborFieldException):
    def __str__(self):
        return (
            f'Analysis field "{self.field}" has been removed '
            f"from arbor field storage in {self.arbor}."
        )


class ArborUnsettableField(ArborFieldException):
    def __str__(self):
        return (
            f'Cannot set values for field "{self.field}" in {self.arbor}. '
            "Only analysis fields can be set."
        )
